@@ -20,6 +20,7 @@
 */
 
 #include "utils.h"
+#include "time.h"
 
 QHostAddress NetworkManager::ipv6AddressAsHostAddress(const QByteArray &address)
 {
@@ -333,6 +334,31 @@ bool NetworkManager::securityIsValid(WirelessSecurityType type, NetworkManager::
             }
         }
         break;
+    case SAE:
+        if (!interfaceCaps.testFlag(NetworkManager::WirelessDevice::Rsn)) {
+            return false;
+        }
+        if (haveAp) {
+            if (!interfaceCaps.testFlag(NetworkManager::WirelessDevice::IBSSRsn)) {
+                return false;
+            }
+            if (apRsn.testFlag(NetworkManager::AccessPoint::PairCcmp) &&
+                    interfaceCaps.testFlag(NetworkManager::WirelessDevice::Ccmp)) {
+                return true;
+            }
+        } else {
+            if (apRsn.testFlag(NetworkManager::AccessPoint::KeyMgmtSAE)) {
+                if (apRsn.testFlag(NetworkManager::AccessPoint::PairTkip) &&
+                        interfaceCaps.testFlag(NetworkManager::WirelessDevice::Tkip)) {
+                    return true;
+                }
+                if (apRsn.testFlag(NetworkManager::AccessPoint::PairCcmp) &&
+                        interfaceCaps.testFlag(NetworkManager::WirelessDevice::Ccmp)) {
+                    return true;
+                }
+            }
+            return false;
+        }
     default:
         good = false;
         break;
@@ -349,7 +375,7 @@ NetworkManager::WirelessSecurityType NetworkManager::findBestWirelessSecurity(Ne
     // Therefore static WEP is before LEAP and Dynamic WEP because there is no way to detect
     // if an AP is capable of Dynamic WEP and showing Dynamic WEP first would confuse
     // Static WEP users.
-    types << NetworkManager::Wpa2Eap << NetworkManager::Wpa2Psk << NetworkManager::WpaEap << NetworkManager::WpaPsk << NetworkManager::StaticWep << NetworkManager::DynamicWep << NetworkManager::Leap << NetworkManager::NoneSecurity;
+    types << NetworkManager::SAE << NetworkManager::Wpa2Eap << NetworkManager::Wpa2Psk << NetworkManager::WpaEap << NetworkManager::WpaPsk << NetworkManager::StaticWep << NetworkManager::DynamicWep << NetworkManager::Leap << NetworkManager::NoneSecurity;
 
     Q_FOREACH (NetworkManager::WirelessSecurityType type, types) {
         if (NetworkManager::securityIsValid(type, interfaceCaps, haveAp, adHoc, apCaps, apWpa, apRsn)) {
@@ -521,3 +547,34 @@ QList<QPair<int, int> > NetworkManager::getAFreqs()
 
     return freqs;
 }
+
+QDateTime NetworkManager::clockBootTimeToDateTime(qlonglong clockBootime) {
+
+    clockid_t clk_id = CLOCK_BOOTTIME;
+    struct timespec tp;
+    int r;
+
+    // now is used as a point of reference
+    // with the timespec that contains the number of msec since boot
+    QDateTime now = QDateTime::currentDateTime();
+    r = clock_gettime (clk_id, &tp);
+    if (r == -1 && errno == EINVAL) {
+        clk_id = CLOCK_MONOTONIC;
+        r = clock_gettime (clk_id, &tp);
+    }
+
+    // convert to msecs
+    long now_msecs = tp.tv_sec * 1000 + tp.tv_nsec / 1000000;
+
+    // diff the msecs and construct a QDateTime based on the offset
+    QDateTime res;
+    if (clockBootime > now_msecs) {
+        qlonglong offset = clockBootime - now_msecs;
+        res = QDateTime::fromMSecsSinceEpoch(now.toMSecsSinceEpoch() + offset);
+    } else {
+        qlonglong offset = now_msecs - clockBootime;
+        res = QDateTime::fromMSecsSinceEpoch(now.toMSecsSinceEpoch() - offset);
+    }
+
+    return res;
+}
